/***************************************************************************
 *   Copyright (C) 2003-2005 by Victor Julien                              *
 *   victor@nk.nl                                                          *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
 
#include "main.h"


/*

	Returncodes:
		0: not filtered
		1: filtered
		
		In case of error we return 0.
*/
static int
filtered_connection(const int debuglvl, struct ConntrackData *cd_ptr, vrcnf_filter *filter)
{
	char	line[512] = "";

	if(cd_ptr == NULL || filter == NULL)
		return(0);

	snprintf(line, sizeof(line), "%d %s %s %s %d %d %d %s %s",
					cd_ptr->cnt,
					cd_ptr->sername,
					cd_ptr->fromname,
					cd_ptr->toname,
					cd_ptr->src_port,
					cd_ptr->dst_port,
					cd_ptr->protocol,
					cd_ptr->src_ip,
					cd_ptr->dst_ip);

	/*	check the regex
		
		If the regex matches, the line is not filtered, so we return 0.
	*/
	if(regexec(&filter->reg, line, 0, NULL, 0) == 0)
	{
		if(filter->neg == FALSE)
			return(0);
		else
			return(1);
	}
	else
	{
		if(filter->neg == FALSE)
			return(1);
		else
			return(0);
	}
}


static int
print_connection(const int debuglvl, WINDOW *local_win, struct ConntrackData *cd_ptr, int draw_by_connect_status, int sort_by_connect_status, int sort_by_in_out_fw, int max_onscreen, int cnt, int screen_width, vrcnf_filter *filter)
{
	int	start_print = 0;
	char	printline[128] = "";
	int	spaceleft = 0,
		printline_width = 0;

	if(filter->reg_active == TRUE)
	{
		if(filtered_connection(debuglvl, cd_ptr, filter) == 1)
			return(0);
	}

	spaceleft = screen_width;

	/* determine the position where we are going to write */
	if(draw_by_connect_status)
	{
		if(cd_ptr->connect_status == CONN_CONNECTING)
		{
			start_print = (max_onscreen/4)*2 + 1 + cnt;
		}
		else if(cd_ptr->connect_status == CONN_DISCONNECTING)
		{
			start_print = (max_onscreen/4)*3 + 1 + cnt;
		}
		else
		{
			start_print = cnt+3;
		}
	}
	else if(sort_by_in_out_fw)
	{
		if(cd_ptr->direction_status == CONN_IN)
		{
			start_print = (max_onscreen / 3) + 3 + 1 + cnt;
		}
		else if(cd_ptr->direction_status == CONN_OUT)
		{
			start_print = (max_onscreen / 3) * 2 + 2 + 1 + cnt;
		}
		else
		{
			start_print = cnt + 3;
		}
	}
	else
	{
		start_print = cnt;
	}

	/*
		count
	*/
	wattron(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);

	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);

	snprintf(printline, printline_width, "%4d: ", cd_ptr->cnt);
	spaceleft = spaceleft - strlen(printline);

	mvwprintw(local_win, start_print, 0, "%s", printline);
	wattroff(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);

	if(!spaceleft)
		return(1);

	/*
		SERVICE name
	*/
	wattron(local_win, COLOR_PAIR(CP_CYAN_BLUE)|A_BOLD);

	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);

	snprintf(printline, printline_width, "%-14s ", cd_ptr->sername);
	spaceleft = spaceleft - strlen(printline);

	wprintw(local_win, "%s", printline);
	wattroff(local_win, COLOR_PAIR(CP_CYAN_BLUE)|A_BOLD);

	if(!spaceleft)
		return(1);

	/*
		FROM name
	*/
	if(strncmp(cd_ptr->fromname, "firewall", 8) == 0)
		wattron(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
	else
		wattron(local_win, COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);

	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);

	snprintf(printline, printline_width, "%-24s ", cd_ptr->fromname);
	spaceleft = spaceleft - strlen(printline);

	wprintw(local_win, "%s", printline);

	
	if(strncmp(cd_ptr->fromname, "firewall", 8) == 0)
		wattroff(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
	else
		wattroff(local_win, COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);

	if(!spaceleft)
		return(1);

	/*
		ARROW
	*/
	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);

	snprintf(printline, printline_width, " -> ");
	spaceleft = spaceleft - strlen(printline);

	wprintw(local_win, "%s", printline);

	if(!spaceleft)
		return(1);

	/*
		TO name
	*/
	if(strncmp(cd_ptr->toname, "firewall", 8) == 0)
		wattron(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
	else
		wattron(local_win, COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);

	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);
	
	snprintf(printline, printline_width, "%-24s ", cd_ptr->toname);
	spaceleft = spaceleft - strlen(printline);

	wprintw(local_win, "%s", printline);
	
	if(strncmp(cd_ptr->toname, "firewall", 8) == 0)
		wattroff(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
	else
		wattroff(local_win, COLOR_PAIR(CP_WHITE_BLUE) | A_BOLD);


	if(!spaceleft)
		return(1);

	/*
		Connection status
	*/
	if(!draw_by_connect_status && sort_by_connect_status)
	{
		/* whitespace */
		printline_width = spaceleft;
		if(printline_width >= sizeof(printline))
			printline_width = sizeof(printline);
	
		snprintf(printline, printline_width, "%s", " ");
		spaceleft = spaceleft - strlen(printline);

		wprintw(local_win, "%s", printline);

		if(cd_ptr->connect_status == CONN_CONNECTING)
		{
			wattron(local_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);
			
			printline_width = spaceleft;
			if(printline_width >= sizeof(printline))
				printline_width = sizeof(printline);
	
			snprintf(printline, printline_width, "%-14s", gettext("CONNECTING"));
			spaceleft = spaceleft - strlen(printline);

			wprintw(local_win, "%s", printline);

			wattroff(local_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);
		}
		else if(cd_ptr->connect_status == CONN_CONNECTED)
		{
			wattron(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);

			printline_width = spaceleft;
			if(printline_width >= sizeof(printline))
				printline_width = sizeof(printline);
	
			snprintf(printline, printline_width, "%-14s", gettext("CONNECTED"));
			spaceleft = spaceleft - strlen(printline);

			wprintw(local_win, "%s", printline);

			wattroff(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
		}
		else if(cd_ptr->connect_status == CONN_DISCONNECTING)
		{
			wattron(local_win, COLOR_PAIR(CP_RED_BLUE) | A_BOLD);
			
			printline_width = spaceleft;
			if(printline_width >= sizeof(printline))
				printline_width = sizeof(printline);
	
			snprintf(printline, printline_width, "%-14s", gettext("DISCONNECTING"));
			spaceleft = spaceleft - strlen(printline);

			wprintw(local_win, "%s", printline);

			wattroff(local_win, COLOR_PAIR(CP_RED_BLUE) | A_BOLD);
		}
		else
			wprintw(local_win, "%-14s", " ");

	}

	if(!spaceleft)
		return(1);

	if(!sort_by_in_out_fw)
	{
		/* whitespace */
		printline_width = spaceleft;
		if(printline_width >= sizeof(printline))
			printline_width = sizeof(printline);
	
		snprintf(printline, printline_width, "%s", " ");
		spaceleft = spaceleft - strlen(printline);

		wprintw(local_win, "%s", printline);

		if(cd_ptr->direction_status == CONN_IN)
		{
			wattron(local_win, COLOR_PAIR(CP_CYAN_BLUE)|A_BOLD);

			printline_width = spaceleft;
			if(printline_width >= sizeof(printline))
				printline_width = sizeof(printline);
	
			snprintf(printline, printline_width, "%-12s", gettext("INCOMING"));
			spaceleft = spaceleft - strlen(printline);

			wprintw(local_win, "%s", printline);
			wattroff(local_win, COLOR_PAIR(CP_CYAN_BLUE) | A_BOLD);
		}
		else if(cd_ptr->direction_status == CONN_OUT)
		{
			wattron(local_win, COLOR_PAIR(CP_CYAN_BLUE)|A_BOLD);

			printline_width = spaceleft;
			if(printline_width >= sizeof(printline))
				printline_width = sizeof(printline);
	
			snprintf(printline, printline_width, "%-12s", gettext("OUTGOING"));
			spaceleft = spaceleft - strlen(printline);

			wprintw(local_win, "%s", printline);
			wattroff(local_win, COLOR_PAIR(CP_CYAN_BLUE) | A_BOLD);
		}
		else if(cd_ptr->direction_status == CONN_FW)
		{
			wattron(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);

			printline_width = spaceleft;
			if(printline_width >= sizeof(printline))
				printline_width = sizeof(printline);
	
			snprintf(printline, printline_width, "%-12s", gettext("FORWARDING"));
			spaceleft = spaceleft - strlen(printline);

			wprintw(local_win, "%s", printline);
			wattroff(local_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
		}
	}


	if(!spaceleft)
		return(1);

	/*
		SERVICE data
	*/
	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);

	snprintf(printline, printline_width, "%5d->%-5d (%2d) ", cd_ptr->src_port, cd_ptr->dst_port, cd_ptr->protocol);

	spaceleft = spaceleft - strlen(printline);
	wprintw(local_win, "%s", printline);

	if(!spaceleft)
		return(1);

	/* FROM ip */
	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);

	snprintf(printline, printline_width, "%-16s", cd_ptr->src_ip);

	spaceleft = spaceleft - strlen(printline);
	wprintw(local_win, "%s", printline);

	if(!spaceleft)
		return(1);

	/* ARROW */
	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);

	snprintf(printline, printline_width, " -> ");
	spaceleft = spaceleft - strlen(printline);

	wprintw(local_win, "%s", printline);

	if(!spaceleft)
		return(1);

	/* TO ip */
	printline_width = spaceleft;
	if(printline_width >= sizeof(printline))
		printline_width = sizeof(printline);
	
	snprintf(printline, printline_width, "%-16s", cd_ptr->dst_ip);

	spaceleft = spaceleft - strlen(printline);
	wprintw(local_win, "%s", printline);

	return(1);
}


int
connections_section(const int debuglvl, Zones *zones, Interfaces *interfaces, Services *services)
{
	int			retval=0;
	WINDOW			*conn_win=NULL;
	PANEL			*my_panels[1];
	int			quit=0,
				ch=0;

	int			max_onscreen=0,
				max_height=0,
				max_width=0,
				connecting=0,
				connected=0,
				disconnecting=0,
				max_connecting=0,
				max_connected=0,
				max_disconnecting=0,
				incoming=0,
				forwarding=0,
				outgoing=0,
				max_incoming=0,
				max_forwarding=0,
				max_outgoing=0;

	/* hashes for the vuurmuur names */
	Hash			ZonedataHash,
				ServicesHash;

	d_list			network_list;

	d_list			Conn_dlist;
	d_list_node		*d_node = NULL;
	struct ConntrackData	*cd_ptr = NULL;
	struct ConntrackStats_	cs;
	int			prev_list_size = 0;

	/* options */
	char			unknown_hosts_as_network = 1,
				sort_by_connect_status = 1,
				sort_by_in_out_fw = 0,
				draw_by_connect_status = 0;

	struct
	{
		char print; /* do we print to screen this run? */
		char sleep; /* do we sleep this run */
		char pause; /* are we in pause mode? 0 no, 1 yes */
	}
	control =
	{
		0,
		0,
		0,
	};


	int	update_interval = 1000000; /* weird, in pratice this seems to be two sec */
	int	slept_so_far    = 1000000; /* time slept since last update */

	/* top menu */
	char	*key_choices[] = 	{	"F12",
						"i",
						"s",
						"c",
						"u",
						"f",
						"F10"};
	int	key_choices_n = 6;
	char	*cmd_choices[] = 	{	gettext("help"),
						gettext("in/out/fw"),
						gettext("status"),
						gettext("connect"),
						gettext("grp unknown"),
						gettext("filter"),
						gettext("back")};
	int	cmd_choices_n = 6;

	vrcnf_filter	vfilter;
	char		use_filter = FALSE;
	int		printed = 0;

	/* init filter */
	filter_setup(debuglvl, &vfilter);

	/* set up & create the logwin */
	getmaxyx(stdscr, max_height, max_width);
	max_onscreen = max_height-8;
	conn_win = newwin(max_height-8, max_width-2, 4, 1);
	wbkgd(conn_win, COLOR_PAIR(3));
	my_panels[0] = new_panel(conn_win);
	keypad(conn_win, TRUE);

	/* make sure wgetch doesn't block the printing of the screen */
	nodelay(conn_win, TRUE);

	/* dont display the cursor */
	curs_set(0);


	/* insert the interfaces as TYPE_FIREWALL's into the zonelist as 'firewall', so this appears in to the connections */
	if(ins_iface_into_zonelist(debuglvl, &interfaces->list, &zones->list) < 0)
	{
		(void)vrprint.error(-1, VR_INTERR, "ins_iface_into_zonelist() failed (in: %s:%d).", __FUNC__, __LINE__);
		return(-1);
	}

	/* do the same for broadcasts. These are removed by: rem_iface_from_zonelist() (see below) */
	if(add_broadcasts_zonelist(debuglvl, zones) < 0)
	{
		(void)vrprint.error(-1, VR_INTERR, "add_broadcasts_zonelist() failed (in: %s:%d).", __FUNC__, __LINE__);
		return(-1);
	}

	/* create hashtables */
	if(init_zonedata_hashtable(debuglvl, zones->list.len * 3, &zones->list, hash_ipaddress, compare_ipaddress, &ZonedataHash) < 0)
	{
		(void)vrprint.error(-1, VR_INTERR, "init_zonedata_hashtable() failed (in: %s:%d).", __FUNC__, __LINE__);
		return(-1);
	}
	/* the hashtable size may seem very big, but some services have really a lot items. e.g. 137->1024:65535 */
	if(init_services_hashtable(debuglvl, services->list.len * 500, &services->list, hash_port, compare_ports, &ServicesHash) < 0)
	{
		(void)vrprint.error(-1, VR_INTERR, "init_services_hashtable() failed (in: %s:%d).", __FUNC__, __LINE__);
		return(-1);
	}

	/* initialize this list with destroy is null, because it only points to zonedatalist nodes */
	if(d_list_setup(debuglvl, &network_list, NULL) < 0)
	{
		(void)vrprint.error(-1, VR_INTERR, "d_list_setup() failed (in: %s:%d).", __FUNC__, __LINE__);
		return(-1);
	}
	if(zonelist_to_networklist(debuglvl, zones, &network_list) < 0)
	{
		(void)vrprint.error(-1, VR_INTERR, "zonelist_to_networklist() failed (in: %s:%d).", __FUNC__, __LINE__);
		return(-1);
	}

	draw_top_menu(debuglvl, top_win, gettext("Connections"), key_choices_n, key_choices, cmd_choices_n, cmd_choices);

	/* the main loop */
	while(quit == 0)
	{
		control.sleep = 1;

		if(control.pause)
			control.print = 0;
		else
			control.print = 1;

		/* check if we have slept long enough */
		if(slept_so_far >= update_interval && !control.pause)
		{
			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNC__, "now update: slept_so_far '%d'.", slept_so_far);

			/* reset the wait counter */
			slept_so_far = 0;

			if(d_list_setup(debuglvl, &Conn_dlist, NULL) < 0)
			{
				(void)vrprint.error(-1, VR_INTERR, "d_list_setup() failed (in: %s:%d).", __FUNC__, __LINE__);
				return(-1);
			}

			/* get the connections from the proc */
			if(conn_get_connections(debuglvl, prev_list_size, &ServicesHash, &ZonedataHash, &Conn_dlist, &network_list, unknown_hosts_as_network, sort_by_connect_status, sort_by_in_out_fw, &cs, NULL, NULL) < 0)
			{
				(void)vrprint.error(-1, VR_ERR, gettext("getting the connections failed."));
				return(-1);
			}

			/* determine how many lines we can draw for each section */
			if(draw_by_connect_status)
			{
				/* connected get half the screen, connecting and disconnecting both 1/4 */
				max_connecting =    (max_onscreen / 4) - 1;
				max_connected =     (max_onscreen / 4) * 2 - 1 - 2;
				max_disconnecting = (max_onscreen / 4) - 1;

				connecting    = 0;
				connected     = 0;
				disconnecting = 0;
			}
			else if(sort_by_in_out_fw)
			{
				/* three equal parts */
				max_incoming =   (max_onscreen / 3) - 2;
				max_forwarding = (max_onscreen / 3);
				max_outgoing =   (max_onscreen / 3) - 3;

				incoming   = 0;
				forwarding = 0;
				outgoing   = 0;
			}

			/* clear screen */
			if(control.print)
				werase(conn_win);

			if(control.print)
			{
				for(printed=0, d_node = Conn_dlist.top; d_node && printed < max_onscreen; d_node = d_node->next)
				{
					if(!(cd_ptr = d_node->data))
					{
						(void)vrprint.error(-1, VR_INTERR, "NULL pointer (in: %s:%d).", __FUNC__, __LINE__);
						return(-1);
					}

					if(draw_by_connect_status)
					{
						if(cd_ptr->connect_status == CONN_CONNECTING)
						{
							if(connecting < max_connecting)
							{
								if(print_connection(debuglvl, conn_win, cd_ptr, draw_by_connect_status, sort_by_connect_status, sort_by_in_out_fw, max_onscreen, connecting, max_width-2, &vfilter) == 1)
								{
									connecting++;
									printed++;
								}
							}
						}
						else if(cd_ptr->connect_status == CONN_DISCONNECTING)
						{
							if(disconnecting < max_disconnecting)
							{
								if(print_connection(debuglvl, conn_win, cd_ptr, draw_by_connect_status, sort_by_connect_status, sort_by_in_out_fw, max_onscreen, disconnecting, max_width-2, &vfilter) == 1)
								{
									disconnecting++;
									printed++;
								}
							}
						}
						else
						{
							if(connected < max_connected)
							{
								if(print_connection(debuglvl, conn_win, cd_ptr, draw_by_connect_status, sort_by_connect_status, sort_by_in_out_fw, max_onscreen, connected, max_width - 2, &vfilter) == 1)
								{
									connected++;
									printed++;
								}
							}
						}

						/* check if it usefull to continue the loop */
						if(	connecting    == max_connecting &&
							disconnecting == max_disconnecting &&
							connected     == max_connected)
						{
							break;
						}
					}
					else if(sort_by_in_out_fw)
					{
						if(cd_ptr->direction_status == CONN_IN)
						{
							if(incoming < max_incoming)
							{
								if(print_connection(debuglvl, conn_win, cd_ptr, draw_by_connect_status, sort_by_connect_status, sort_by_in_out_fw, max_onscreen, incoming, max_width - 2, &vfilter) == 1)
								{
									incoming++;
									printed++;
								}
							}
						}
						else if(cd_ptr->direction_status == CONN_OUT)
						{
							if(outgoing < max_outgoing)
							{
								if(print_connection(debuglvl, conn_win, cd_ptr, draw_by_connect_status, sort_by_connect_status, sort_by_in_out_fw, max_onscreen, outgoing, max_width - 2, &vfilter) == 1)
								{
									outgoing++;
									printed++;
								}
							}
						}
						else
						{
							if(forwarding < max_forwarding)
							{
								if(print_connection(debuglvl, conn_win, cd_ptr, draw_by_connect_status, sort_by_connect_status, sort_by_in_out_fw, max_onscreen, forwarding, max_width - 2, &vfilter) == 1)
								{
									forwarding++;
									printed++;
								}
							}
						}

						/* check if it usefull to continue the loop */
						if(	incoming   == max_incoming &&
							outgoing   == max_outgoing &&
							forwarding == max_forwarding)
						{
							break;
						}
					}
					else
					{
						if(print_connection(debuglvl, conn_win, cd_ptr, draw_by_connect_status, sort_by_connect_status, sort_by_in_out_fw, max_onscreen, printed, max_width - 2, &vfilter) == 1)
						{
							printed++;
						}

						/* check if it usefull to continue the loop */
						if(printed == max_onscreen)
							break;
					}
				}

				/* print the seperators */
				if(draw_by_connect_status)
				{
					wattron(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);
					mvwprintw(conn_win, 0, 4, "%s:",  gettext("Connections"));
					mvwprintw(conn_win, 0, 20, "%s:", gettext("Total"));
					mvwprintw(conn_win, 0, 40, "%s:", gettext("Incoming"));
					mvwprintw(conn_win, 0, 64, "%s:", gettext("Forwarding"));
					mvwprintw(conn_win, 1, 40, "%s:", gettext("Outgoing"));

					mvwprintw(conn_win, 0, 34, "%4d", cs.conn_total);
					mvwprintw(conn_win, 0, 58, "%4d", cs.conn_in);
					mvwprintw(conn_win, 0, 78, "%4d", cs.conn_fw);
					mvwprintw(conn_win, 1, 58, "%4d", cs.conn_out);

					wattroff(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);

					//
					mvwhline(conn_win, 2, 0, ACS_HLINE, max_width-2);
					wattron(conn_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
					mvwprintw(conn_win, 2, 3, " %s ", gettext("Established Connections"));
					mvwprintw(conn_win, 2, max_width-11, " (%d) ", cs.stat_estab);
					wattroff(conn_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);

					// print at the half of the screen
					mvwhline(conn_win, (max_onscreen/4)*2, 0, ACS_HLINE, max_width-2);
					wattron(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);
					mvwprintw(conn_win, (max_onscreen/4)*2, 3, " %s ", gettext("Connections Initializing"));
					mvwprintw(conn_win, (max_onscreen/4)*2, max_width-11, " (%d) ", cs.stat_connect);
					wattroff(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);

					mvwhline(conn_win, (max_onscreen/4)*3, 0, ACS_HLINE, max_width-2);
					wattron(conn_win, COLOR_PAIR(CP_RED_BLUE) | A_BOLD);
					mvwprintw(conn_win, (max_onscreen/4)*3, 3, " %s ", gettext("Connections Closing"));
					mvwprintw(conn_win, (max_onscreen/4)*3, max_width-11, " (%d) ", cs.stat_closing);
					wattroff(conn_win, COLOR_PAIR(CP_RED_BLUE) | A_BOLD);

					// move the cursor a bit out of sight
					mvwprintw(conn_win, max_onscreen-1, max_width-3, " ");
				}

				if(sort_by_in_out_fw)
				{
					wattron(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);
					mvwprintw(conn_win, 0, 4, "%s:",  gettext("Connections"));
					mvwprintw(conn_win, 0, 20, "%s:", gettext("Total"));
					mvwprintw(conn_win, 0, 40, "%s:", gettext("Connecting"));
					mvwprintw(conn_win, 0, 64, "%s:", gettext("Established"));
					mvwprintw(conn_win, 1, 40, "%s:", gettext("Disconnecting"));

					mvwprintw(conn_win, 0, 34, "%4d", cs.conn_total);
					mvwprintw(conn_win, 0, 58, "%4d", cs.stat_connect);
					mvwprintw(conn_win, 0, 78, "%4d", cs.stat_estab);
					mvwprintw(conn_win, 1, 58, "%4d", cs.stat_closing);
					wattroff(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);

					/* */
					mvwhline(conn_win, 2, 0, ACS_HLINE, max_width-2);
					wattron(conn_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);
					mvwprintw(conn_win, 2, 3, " %s ", gettext("Forwarded Connections"));
					mvwprintw(conn_win, 2, max_width-11, " (%d) ", cs.conn_fw);
					wattroff(conn_win, COLOR_PAIR(CP_YELLOW_BLUE) | A_BOLD);

					/* print at the one third of the screen */
					mvwhline(conn_win, (max_onscreen/3)+3, 0, ACS_HLINE, max_width-2);
					wattron(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);
					mvwprintw(conn_win, (max_onscreen/3)+3, 3, " %s ", gettext("Incoming Connections"));
					mvwprintw(conn_win, (max_onscreen/3)+3, max_width-11, " (%d) ", cs.conn_in);
					wattroff(conn_win, COLOR_PAIR(CP_GREEN_BLUE) | A_BOLD);

					mvwhline(conn_win, (max_onscreen/3)*2+2, 0, ACS_HLINE, max_width-2);
					wattron(conn_win, COLOR_PAIR(CP_RED_BLUE) | A_BOLD);
					mvwprintw(conn_win, (max_onscreen/3)*2+2, 3, " %s ", gettext("Outgoing Connections"));
					mvwprintw(conn_win, (max_onscreen/3)*2+2, max_width-11, " (%d) ", cs.conn_out);
					wattroff(conn_win, COLOR_PAIR(CP_RED_BLUE) | A_BOLD);

					/* move the cursor a bit out of sight */
					mvwprintw(conn_win, max_onscreen-1, max_width-3, " ");
				}
		
				wrefresh(conn_win);
			}

			/* store prev list size */
			prev_list_size = Conn_dlist.len;

			/* clean up the list */
			conn_list_cleanup(debuglvl, &Conn_dlist);
		}


		/*
			//////// HANDLE KEYBOARD //////////
		*/
		ch = wgetch(conn_win);
		switch(ch)
		{
			/* QUIT */
			case 'q':
			case 'Q':
			case KEY_F(10):
				quit = 1;
				control.sleep = 0;
				break;

			case 'u':
				if(unknown_hosts_as_network == 1)
					unknown_hosts_as_network = 0;
				else
					unknown_hosts_as_network = 1;

				control.sleep = 0;
				break;

			case 's':
				if(sort_by_connect_status == 1)
					sort_by_connect_status = 0;
				else
					sort_by_connect_status = 1;

				control.sleep = 0;
				break;

			case 'i':
				if(sort_by_in_out_fw == 1)
					sort_by_in_out_fw = 0;
				else
				{
					sort_by_in_out_fw = 1;
					draw_by_connect_status = 0;
				}

				control.sleep = 0;
				break;

			case 'c':
				if(draw_by_connect_status == 1)
					draw_by_connect_status = 0;
				else
				{
					draw_by_connect_status = 1;
					sort_by_in_out_fw = 0;
					sort_by_connect_status = 1;
				}

				control.sleep = 0;
				break;

			// PAUSE
			case 'p':
			case 32: // space
				if(control.pause == 1)
				{
					status_print(status_win, "");
					control.pause = 0;
				}
				else
				{
					control.pause = 1;
					status_print(status_win, "*** PAUSED *** (press 'p' to continue)");
				}

				control.sleep = 0;
				break;

			case 'f':
			case 'F':
			case 10:

				if(ch != 10)
				{
					filter_input_box(debuglvl, &vfilter);
				}
				else
				{
					filter_cleanup(debuglvl, &vfilter);
				}

				if(vfilter.reg_active == TRUE)
				{
					status_print(status_win, gettext("Active filter: '%s' (press 'enter' to clear)."), vfilter.str);
					use_filter = TRUE;
				}
				else if(use_filter == TRUE && vfilter.reg_active == FALSE)
				{
					status_print(status_win, gettext("Filter removed."));
					use_filter = FALSE;
				}

				break;

			case KEY_F(12):
			case 'h':
			case 'H':
			case '?':
				print_help(debuglvl, ":[VUURMUUR:CONNECTIONS]:");
				break;
		}


		/* now sleep! */
		if(control.sleep == 1)
		{
			usleep(10000);
			slept_so_far = slept_so_far + 10000;

			//(void)vrprint.debug(__FUNC__, "just slept: slept_so_far '%d'.", slept_so_far);
		}
		else
		{
			slept_so_far = update_interval;

			if(debuglvl >= LOW)
				(void)vrprint.debug(__FUNC__, "control.sleep = 0: set slept_so_far to update_interval.");
		}
	}


	/* EXIT: cleanup */
	d_list_cleanup(debuglvl, &network_list);

	/*	remove the interfaces inserted as TYPE_FIREWALL's into the zonelist
		this also removes zones added by add_broadcasts_zonelist()
	*/
	if(rem_iface_from_zonelist(debuglvl, &zones->list) < 0)
	{
		(void)vrprint.error(-1, VR_INTERR, "rem_iface_from_zonelist() failed (in: %s:%d).", __FUNC__, __LINE__);
		retval = -1;
	}

	/* filter clean up */
	filter_cleanup(debuglvl, &vfilter);

	/*
		destroy hashtables
	*/
	hash_cleanup(debuglvl, &ZonedataHash);
	hash_cleanup(debuglvl, &ServicesHash);

	nodelay(conn_win, FALSE);

	del_panel(my_panels[0]);
	destroy_win(conn_win);

	/* display the cursor again */
	curs_set(1);

	update_panels();
	doupdate();

	return(retval);
}
